@@ -3,31 +3,50 @@
 This is conceptually part of mypy.semanal (semantic analyzer pass 2).
 """
 
-from typing import Tuple, Optional
+from __future__ import annotations
 
-from mypy.types import (
-    Type, Instance, CallableType, NoneType, TupleType, AnyType, PlaceholderType,
-    TypeOfAny, get_proper_type
-)
-from mypy.nodes import (
-    AssignmentStmt, NewTypeExpr, CallExpr, NameExpr, RefExpr, Context, StrExpr, BytesExpr,
-    UnicodeExpr, Block, FuncDef, Argument, TypeInfo, Var, SymbolTableNode, MDEF, ARG_POS,
-    PlaceholderNode
-)
-from mypy.semanal_shared import SemanticAnalyzerInterface
-from mypy.options import Options
-from mypy.exprtotype import expr_to_unanalyzed_type, TypeTranslationError
-from mypy.typeanal import check_for_explicit_any, has_any_from_unimported_type
-from mypy.messages import MessageBuilder, format_type
-from mypy.errorcodes import ErrorCode
 from mypy import errorcodes as codes
+from mypy.errorcodes import ErrorCode
+from mypy.exprtotype import TypeTranslationError, expr_to_unanalyzed_type
+from mypy.messages import MessageBuilder, format_type
+from mypy.nodes import (
+    ARG_POS,
+    MDEF,
+    Argument,
+    AssignmentStmt,
+    Block,
+    CallExpr,
+    Context,
+    FuncDef,
+    NameExpr,
+    NewTypeExpr,
+    PlaceholderNode,
+    RefExpr,
+    StrExpr,
+    SymbolTableNode,
+    TypeInfo,
+    Var,
+)
+from mypy.options import Options
+from mypy.semanal_shared import SemanticAnalyzerInterface, has_placeholder
+from mypy.typeanal import check_for_explicit_any, has_any_from_unimported_type
+from mypy.types import (
+    AnyType,
+    CallableType,
+    Instance,
+    NoneType,
+    PlaceholderType,
+    TupleType,
+    Type,
+    TypeOfAny,
+    get_proper_type,
+)
 
 
 class NewTypeAnalyzer:
-    def __init__(self,
-                 options: Options,
-                 api: SemanticAnalyzerInterface,
-                 msg: MessageBuilder) -> None:
+    def __init__(
+        self, options: Options, api: SemanticAnalyzerInterface, msg: MessageBuilder
+    ) -> None:
         self.options = options
         self.api = api
         self.msg = msg
@@ -50,19 +69,20 @@ class NewTypeAnalyzer:
         # add placeholder as we do for ClassDef.
 
         if self.api.is_func_scope():
-            name += '@' + str(s.line)
+            name += "@" + str(s.line)
         fullname = self.api.qualified_name(name)
 
-        if (not call.analyzed or
-                isinstance(call.analyzed, NewTypeExpr) and not call.analyzed.info):
+        if not call.analyzed or isinstance(call.analyzed, NewTypeExpr) and not call.analyzed.info:
             # Start from labeling this as a future class, as we do for normal ClassDefs.
             placeholder = PlaceholderNode(fullname, s, s.line, becomes_typeinfo=True)
             self.api.add_symbol(var_name, placeholder, s, can_defer=False)
 
         old_type, should_defer = self.check_newtype_args(var_name, call, s)
         old_type = get_proper_type(old_type)
-        if not call.analyzed:
+        if not isinstance(call.analyzed, NewTypeExpr):
             call.analyzed = NewTypeExpr(var_name, old_type, line=call.line, column=call.column)
+        else:
+            call.analyzed.old_type = old_type
         if old_type is None:
             if should_defer:
                 # Base type is not ready.
@@ -70,26 +90,33 @@ class NewTypeAnalyzer:
                 return True
 
         # Create the corresponding class definition if the aliased type is subtypeable
+        assert isinstance(call.analyzed, NewTypeExpr)
         if isinstance(old_type, TupleType):
-            newtype_class_info = self.build_newtype_typeinfo(name, old_type,
-                                                             old_type.partial_fallback, s.line)
-            newtype_class_info.tuple_type = old_type
+            newtype_class_info = self.build_newtype_typeinfo(
+                name, old_type, old_type.partial_fallback, s.line, call.analyzed.info
+            )
+            newtype_class_info.update_tuple_type(old_type)
         elif isinstance(old_type, Instance):
             if old_type.type.is_protocol:
                 self.fail("NewType cannot be used with protocol classes", s)
-            newtype_class_info = self.build_newtype_typeinfo(name, old_type, old_type, s.line)
+            newtype_class_info = self.build_newtype_typeinfo(
+                name, old_type, old_type, s.line, call.analyzed.info
+            )
         else:
             if old_type is not None:
                 message = "Argument 2 to NewType(...) must be subclassable (got {})"
                 self.fail(message.format(format_type(old_type)), s, code=codes.VALID_NEWTYPE)
             # Otherwise the error was already reported.
             old_type = AnyType(TypeOfAny.from_error)
-            object_type = self.api.named_type('__builtins__.object')
-            newtype_class_info = self.build_newtype_typeinfo(name, old_type, object_type, s.line)
+            object_type = self.api.named_type("builtins.object")
+            newtype_class_info = self.build_newtype_typeinfo(
+                name, old_type, object_type, s.line, call.analyzed.info
+            )
             newtype_class_info.fallback_to_any = True
 
-        check_for_explicit_any(old_type, self.options, self.api.is_typeshed_stub_file, self.msg,
-                               context=s)
+        check_for_explicit_any(
+            old_type, self.options, self.api.is_typeshed_stub_file, self.msg, context=s
+        )
 
         if self.options.disallow_any_unimported and has_any_from_unimported_type(old_type):
             self.msg.unimported_type_becomes_any("Argument 2 to NewType(...)", old_type, s)
@@ -108,15 +135,16 @@ class NewTypeAnalyzer:
         newtype_class_info.line = s.line
         return True
 
-    def analyze_newtype_declaration(self,
-            s: AssignmentStmt) -> Tuple[Optional[str], Optional[CallExpr]]:
+    def analyze_newtype_declaration(self, s: AssignmentStmt) -> tuple[str | None, CallExpr | None]:
         """Return the NewType call expression if `s` is a newtype declaration or None otherwise."""
         name, call = None, None
-        if (len(s.lvalues) == 1
-                and isinstance(s.lvalues[0], NameExpr)
-                and isinstance(s.rvalue, CallExpr)
-                and isinstance(s.rvalue.callee, RefExpr)
-                and s.rvalue.callee.fullname == 'typing.NewType'):
+        if (
+            len(s.lvalues) == 1
+            and isinstance(s.lvalues[0], NameExpr)
+            and isinstance(s.rvalue, CallExpr)
+            and isinstance(s.rvalue.callee, RefExpr)
+            and s.rvalue.callee.fullname == "typing.NewType"
+        ):
             name = s.lvalues[0].name
 
             if s.type:
@@ -125,9 +153,12 @@ class NewTypeAnalyzer:
             names = self.api.current_symbol_table()
             existing = names.get(name)
             # Give a better error message than generic "Name already defined".
-            if (existing and
-                    not isinstance(existing.node, PlaceholderNode) and not s.rvalue.analyzed):
-                self.fail('Cannot redefine "%s" as a NewType' % name, s)
+            if (
+                existing
+                and not isinstance(existing.node, PlaceholderNode)
+                and not s.rvalue.analyzed
+            ):
+                self.fail(f'Cannot redefine "{name}" as a NewType', s)
 
             # This dummy NewTypeExpr marks the call as sufficiently analyzed; it will be
             # overwritten later with a fully complete NewTypeExpr if there are no other
@@ -136,8 +167,9 @@ class NewTypeAnalyzer:
 
         return name, call
 
-    def check_newtype_args(self, name: str, call: CallExpr,
-                           context: Context) -> Tuple[Optional[Type], bool]:
+    def check_newtype_args(
+        self, name: str, call: CallExpr, context: Context
+    ) -> tuple[Type | None, bool]:
         """Ananlyze base type in NewType call.
 
         Return a tuple (type, should defer).
@@ -149,7 +181,7 @@ class NewTypeAnalyzer:
             return None, False
 
         # Check first argument
-        if not isinstance(args[0], (StrExpr, BytesExpr, UnicodeExpr)):
+        if not isinstance(args[0], StrExpr):
             self.fail("Argument 1 to NewType(...) must be a string literal", context)
             has_failed = True
         elif args[0].value != name:
@@ -160,17 +192,25 @@ class NewTypeAnalyzer:
         # Check second argument
         msg = "Argument 2 to NewType(...) must be a valid type"
         try:
-            unanalyzed_type = expr_to_unanalyzed_type(args[1])
+            unanalyzed_type = expr_to_unanalyzed_type(args[1], self.options, self.api.is_stub_file)
         except TypeTranslationError:
             self.fail(msg, context)
             return None, False
 
         # We want to use our custom error message (see above), so we suppress
         # the default error message for invalid types here.
-        old_type = get_proper_type(self.api.anal_type(unanalyzed_type,
-                                                      report_invalid_types=False))
+        old_type = get_proper_type(
+            self.api.anal_type(
+                unanalyzed_type,
+                report_invalid_types=False,
+                allow_placeholder=not self.options.disable_recursive_aliases
+                and not self.api.is_func_scope(),
+            )
+        )
         should_defer = False
-        if old_type is None or isinstance(old_type, PlaceholderType):
+        if isinstance(old_type, PlaceholderType):
+            old_type = None
+        if old_type is None:
             should_defer = True
 
         # The caller of this function assumes that if we return a Type, it's always
@@ -181,26 +221,38 @@ class NewTypeAnalyzer:
 
         return None if has_failed else old_type, should_defer
 
-    def build_newtype_typeinfo(self, name: str, old_type: Type, base_type: Instance,
-                               line: int) -> TypeInfo:
-        info = self.api.basic_new_typeinfo(name, base_type, line)
+    def build_newtype_typeinfo(
+        self,
+        name: str,
+        old_type: Type,
+        base_type: Instance,
+        line: int,
+        existing_info: TypeInfo | None,
+    ) -> TypeInfo:
+        info = existing_info or self.api.basic_new_typeinfo(name, base_type, line)
+        info.bases = [base_type]  # Update in case there were nested placeholders.
         info.is_newtype = True
 
         # Add __init__ method
-        args = [Argument(Var('self'), NoneType(), None, ARG_POS),
-                self.make_argument('item', old_type)]
+        args = [
+            Argument(Var("self"), NoneType(), None, ARG_POS),
+            self.make_argument("item", old_type),
+        ]
         signature = CallableType(
             arg_types=[Instance(info, []), old_type],
             arg_kinds=[arg.kind for arg in args],
-            arg_names=['self', 'item'],
+            arg_names=["self", "item"],
             ret_type=NoneType(),
-            fallback=self.api.named_type('__builtins__.function'),
-            name=name)
-        init_func = FuncDef('__init__', args, Block([]), typ=signature)
+            fallback=self.api.named_type("builtins.function"),
+            name=name,
+        )
+        init_func = FuncDef("__init__", args, Block([]), typ=signature)
         init_func.info = info
-        init_func._fullname = info.fullname + '.__init__'
-        info.names['__init__'] = SymbolTableNode(MDEF, init_func)
+        init_func._fullname = info.fullname + ".__init__"
+        info.names["__init__"] = SymbolTableNode(MDEF, init_func)
 
+        if has_placeholder(old_type) or info.tuple_type and has_placeholder(info.tuple_type):
+            self.api.defer(force_progress=True)
         return info
 
     # Helpers
@@ -208,5 +260,5 @@ class NewTypeAnalyzer:
     def make_argument(self, name: str, type: Type) -> Argument:
         return Argument(Var(name), type, None, ARG_POS)
 
-    def fail(self, msg: str, ctx: Context, *, code: Optional[ErrorCode] = None) -> None:
+    def fail(self, msg: str, ctx: Context, *, code: ErrorCode | None = None) -> None:
         self.api.fail(msg, ctx, code=code)
